/**
 * Copyright 2009, AIAN
 * This code may be freely used in any application.
 *
 * ORIGINAL IMPLEMENTATION NOTICE:
 * This hashing class is based on the NIST's Secure Hash Algorithm (FIPS 180).
 * And the original code written in C is implemented by James J. Gillogly as a
 * free-usable code. You can refer to the original at the URL below.
 *
 * http://www.ussrback.com/crypto/source/hash/SHS/t/sha.c
 *
 * Original Copyright 1993, Dr. James J. Gillogly
 */
#include <memory.h>
#include <sstream>
#include <iomanip>
#include "hash.h"

const Hash::U32 Hash::K0 = 0x5a827999;
const Hash::U32 Hash::K1 = 0x6ed9eba1;
const Hash::U32 Hash::K2 = 0x8f1bbcdc;
const Hash::U32 Hash::K3 = 0xca62c1d6;

#define r0(f, K) \
    temp = S(5, A) + f(B, C, D) + E + *p0++ + K; \
    E = D; \
    D = C; \
    C = S(30, B); \
    B = A; \
    A = temp

#define r1(f, K) \
    temp = S(5, A) + f(B, C, D) + E + \
        (*p0++ = *p1++ ^ *p2++ ^ *p3++ ^ *p4++) + K; \
    E = D; \
    D = C; \
    C = S(30, B); \
    B = A; \
    A = temp

Hash::Hash() {
    hbuf_[0] = 0;
    hbuf_[1] = 0;
    hbuf_[2] = 0;
    hbuf_[3] = 0;
    hbuf_[4] = 0;
}

Hash::~Hash() {
}

Hash::U32
Hash::f0(U32 x, U32 y, U32 z) {
    return (z ^ (x & (y ^ z)));
}
Hash::U32
Hash::f1(U32 x, U32 y, U32 z) {
    return (x ^ y ^ z);
}
Hash::U32
Hash::f2(U32 x, U32 y, U32 z) {
    return ((x & y) | (z & (x | y)));
}
Hash::U32
Hash::f3(U32 x, U32 y, U32 z) {
    return (x ^ y ^ z);
}
Hash::U32
Hash::S(U32 n, U32 x) {
    return  ((x << n) | (x >> (32 - n)));
}

void
Hash::doHash(const char* mem, U32 length) {
    LongByte d;
    U32 nread;
    U32 nbits;
    U32 hi_length = 0;
    U32 lo_length = 0;
    U32 h0 = 0x67452301;
    U32 h1 = 0xefcdab89;
    U32 h2 = 0x98badcfe;
    U32 h3 = 0x10325476;
    U32 h4 = 0xc3d2e1f0;
    U32 A, B, C, D, E;
    U32 *p0, *p1, *p2, *p3, *p4;
    U32 temp;

    bool padded = false;
    const char* s = mem;
    while (true) {
        nread = (length < 64) ? length : 64;
        memcpy(d.B, s, nread);
        length -= nread; s += nread;

        if (nread < 64) {
            nbits = nread << 3;
            if ((lo_length += nbits) < nbits) {
                hi_length++;
            }
            if (nread < 64 && !padded) {
                d.B[nread++] = 0x80;
                padded = true;
            }
            for (int i = nread; i < 64; i++) {
                d.B[i] = 0;
            }
            if (nread <= 56) {
                d.W[14] = hi_length;
                d.W[15] = lo_length;
            }
        } else {
            if ((lo_length += 512) < 512) {
                hi_length++;
            }
        }
        p0 = d.W;

        A = h0; B = h1; C = h2; D = h3; E = h4;
        r0(f0,K0); r0(f0,K0); r0(f0,K0); r0(f0,K0); r0(f0,K0);
        r0(f0,K0); r0(f0,K0); r0(f0,K0); r0(f0,K0); r0(f0,K0);
        r0(f0,K0); r0(f0,K0); r0(f0,K0); r0(f0,K0); r0(f0,K0);
        r0(f0,K0);
        p1 = &d.W[13]; p2 = &d.W[8]; p3 = &d.W[2]; p4 = &d.W[0];
        r1(f0,K0); r1(f0,K0); r1(f0,K0); r1(f0,K0);
        r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1);
        r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1);
        r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1);
        r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1); r1(f1,K1);
        r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2);
        r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2);
        r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2);
        r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2); r1(f2,K2);
        r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3);
        r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3);
        r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3);
        r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3); r1(f3,K3);
        h0 += A; h1 += B; h2 += C; h3 += D; h4 += E;

        if (nread <= 56) {
            break;
        }
    }
    hbuf_[0] = h0;
    hbuf_[1] = h1;
    hbuf_[2] = h2;
    hbuf_[3] = h3;
    hbuf_[4] = h4;
}

std::string
Hash::get(const std::string& mem) {
    doHash(mem.c_str(), mem.length());
    std::stringstream ss;
    ss << std::hex << std::setw(8) << std::setfill('0') << hbuf_[0];
    ss << std::hex << std::setw(8) << std::setfill('0') << hbuf_[1];
    ss << std::hex << std::setw(8) << std::setfill('0') << hbuf_[2];
    ss << std::hex << std::setw(8) << std::setfill('0') << hbuf_[3];
    ss << std::hex << std::setw(8) << std::setfill('0') << hbuf_[4];
    return ss.str();
}
